// SpecialNumbers.cpp : Defines the entry point for the console application.
//
/*
This abusive floating point program tries to demonstrate performance
problems, primarily on Intel processors. It tries to demonstrate the
worst possible cost of denormals, NANs, and infinities, and also measures
the divide, square root and multiply performance at various precisions and
with various types of error conditions.

Copyright 2003 Bruce Dawson
*/

#include "stdafx.h"

#include <cstdio>
#include <float.h>
#include <cassert>
#include <cmath>

using namespace std;

#include "support.h"

// Increase this (default is 1,000,000) to run longer tests, for
// slightly better accuracy.
const int kMinCount = 1000000;

// Define this to do tests of sqrt, divide and mul speed at various precisions and
// with various types of data, including data that causes overflow and underflow
#define	DO_FULL_MATH_TESTS

// Maximum possible alignment amount we could care about testing.
// Must be a power of two.
const int kAlignSize = 4096;

// Some useful double values we will be testing with.
double inf = -log(0);
double nan = log(-1);
double nrm = pow(.5, 1022);
double den = nrm * pow(.5, 1);
double zero = 0;

// Useful functions for printing out the results all neat and tidy.
void AddTimings( const char* identifier, int count, double elapsedTime, double baseSpeed );
void AddHeading( const char* heading );

// A handy function/global pair to get our parameter in a guaranteed aligned
// double, so that alignment differences don't interfere with our other
// timing tests.
double g_parameter;
void SetParameter( double parameter )
{
	g_parameter = parameter;
}




double AddTestReg(const char* identifier, double start, double adder, double baseSpeed, int count = kMinCount)
{
	int loops = count / 10;
	__int64 startTime = GetCycleCount();
	__asm
	{
		// Load five copies of start
		push	ecx
		fld		start
		fld		start
		fld		start
		fld		start
		fld		start
		// Load the adder
		fld		adder
		mov		ecx, loops

LoopStart:
		// Do five independent adds, so that the CPU can overlap them if it has
		// that ability.
		fadd	st(1), st(0)
		fadd	st(2), st(0)
		fadd	st(3), st(0)
		fadd	st(4), st(0)
		fadd	st(5), st(0)

		// Do five more adds, to reduce loop overhead.
		fadd	st(1), st(0)
		fadd	st(2), st(0)
		fadd	st(3), st(0)
		fadd	st(4), st(0)
		fadd	st(5), st(0)
		sub		ecx, 1
		jnz    LoopStart

		// Pop the numbers off of the floating point stack.
		fstp	start
		fstp	start
		fstp	start
		fstp	start
		fstp	start
		fstp	start
		pop		ecx
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}




double AddTestMem(const char* identifier, double start, double adder, double baseSpeed, int count = kMinCount)
{
	// Copy the parameter to a global variable in a way that the optimizer will leave
	// it alone. We need to add from a global variable because the alignment of
	// parameters is not guaranteed. A double that is not eight byte aligned incurs a
	// single cycle penalty, but a double that crosses a cache line boundary incurs
	// a massive 10+ cycle delay! Page boundaries are even worse.
	SetParameter( adder );
	int loops = count / 10;
	__int64 startTime = GetCycleCount();
	__asm
	{
		push	ecx
		fld		start
		mov		ecx, loops

LoopStart:
		// These adds are, necessarily, not done in parallel. That's because
		// adds from memory always go to ST(0)
		fadd	g_parameter
		fadd	g_parameter
		fadd	g_parameter
		fadd	g_parameter
		fadd	g_parameter

		fadd	g_parameter
		fadd	g_parameter
		fadd	g_parameter
		fadd	g_parameter
		fadd	g_parameter
		sub		ecx, 1
		jnz    LoopStart

		fstp	start
		pop		ecx
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}




double LoadTest( const char* identifier, double start, double adder,
			int sourceAlignOffset, int destAlignOffset, double baseSpeed, int count = kMinCount)
{
	// Reserve enough memory so that we can find an aligned page within it and still have
	// room beyond it. Really only needs to be kAlignSize + sizeof(double) - 2, but this
	// extra size gives us more flexibility.
	char	alignBuffer[ kAlignSize * 3 ];
	// Find the address of the first page boundary within the buffer.
	char*	page = (char *)((((int)alignBuffer) + (kAlignSize - 1)) & ~( kAlignSize - 1 ) );

	double* source = (double *)( page + sourceAlignOffset );
	*source = start;

	double* dest = (double *)( page + destAlignOffset );
	*dest = adder;

	int loops = count / 5;
	__int64 startTime = GetCycleCount();
	__asm
	{
		push	eax
		push	ebx
		push	ecx
		mov		eax, source
		mov		ebx, dest
		mov		ecx, loops

LoopStart:
		fld		QWORD PTR [eax]
		fld		QWORD PTR [eax]
		fld		QWORD PTR [eax]
		fld		QWORD PTR [eax]
		fld		QWORD PTR [eax]

		fstp	QWORD PTR [ebx]
		fstp	QWORD PTR [ebx]
		fstp	QWORD PTR [ebx]
		fstp	QWORD PTR [ebx]
		fstp	QWORD PTR [ebx]
		sub		ecx, 1
		jnz    LoopStart

		pop		ecx
		pop		ebx
		pop		eax
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}




double DivTest(const char* identifier, double start, double divisor, double baseSpeed, int count = kMinCount)
{
	// Guarantee alignment.
	SetParameter( divisor );
	__int64 startTime = GetCycleCount();
	for ( int i = 0; i < count; ++i )
	{
		SetResult( start / g_parameter );
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}




double MulTest(const char* identifier, double start, double multiplier, double baseSpeed, int count = kMinCount)
{
	// Guarantee alignment.
	SetParameter( multiplier );
	__int64 startTime = GetCycleCount();
	for ( int i = 0; i < count; ++i )
	{
		SetResult( start * g_parameter );
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}




double SqrtTest(const char* identifier, double source, double baseSpeed, int count = kMinCount)
{
	// Guarantee alignment.
	SetParameter( source );
	__int64 startTime = GetCycleCount();
	for ( int i = 0; i < count; ++i )
	{
		SetResult( sqrt( g_parameter ) );
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}


void ADDWithSSE2( double *result, double val0, double val1 )
{
	_asm
	{
		mov eax, result;
		movsd xmm0, val0
		movsd xmm1, val1
		addsd xmm0, xmm1
		movsd [eax], xmm0
	}
}



double MulTestSSE2(const char* identifier, double start, double multiplier, double baseSpeed, int count = kMinCount)
{
	// Guarantee alignment.
	SetParameter( multiplier );
	double result;
	__int64 startTime = GetCycleCount();
    int loopCount = count / 5;
	for ( int i = 0; i < loopCount; ++i )
	{
		_asm
		{
			movsd xmm1, g_parameter
			movsd xmm0, start
			mulsd xmm0, xmm1

			movsd xmm0, start
			mulsd xmm0, xmm1

			movsd xmm0, start
			mulsd xmm0, xmm1

			movsd xmm0, start
			mulsd xmm0, xmm1

			movsd xmm0, start
			mulsd xmm0, xmm1
            movsd result, xmm0
		}
		SetResult( result );
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}



double AddTestSSE2(const char* identifier, double param1, double param2, double baseSpeed, int count = kMinCount)
{
	//double result;
	__int64 startTime = GetCycleCount();
	for ( int i = 0; i < count; ++i )
	{
		_asm
		{
			movsd xmm0, param1
			movsd xmm1, param2
			addsd xmm0, xmm1
			//movsd result, xmm0
		}
		//SetResult( result );
	}
	double elapsedTime = double(GetCycleCount() - startTime) / GetCPUSpeed();

	AddTimings( identifier, count, elapsedTime, baseSpeed );

	return count / elapsedTime;
}




int main(int argc, char* argv[])
{
	assert( (_fpclass( inf ) == _FPCLASS_PINF) );	// Make sure we are Positive INFinity
	assert( (_fpclass( nan ) == _FPCLASS_QNAN) );	// Make sure we are Quiet NAN
	assert( (_fpclass( nrm ) == _FPCLASS_PN) );		// Make sure we are Positive Normalized
	assert( (_fpclass( den ) == _FPCLASS_PD) );		// Make sure we are Positive Denormalized
	assert( (_fpclass( zero ) == _FPCLASS_PZ) );	// Make sure we are Positive Zero

	// Redundant asserts
	assert( !_finite(inf) && !_isnan(inf) );
	assert( _isnan(nan) );

	// Find out how fast our CPU is - will crash on CPUs that don't support RDTSC
	GetCPUSpeed();

	const bool sse2 = HasSSE2();
	if (sse2)
	{
		printf( "CPU supports the SSE2 instruction set\n" );

		// Test denormalized number processing
		double result;
		ADDWithSSE2( &result, den, den );
		if (result == 0)
			printf( "Processing denormalized numbers with SSE2 forces them to zero\n" );
		else
			printf( "SSE2 Denormalized processing seems to be correct\n" );
	}
	else
		printf( "CPU does not support the SSE2 instruction set\n" );

	// Start our tests.

#ifdef	DO_FULL_MATH_TESTS	
	AddHeading( "Divide tests - float precision" );
	_controlfp( _PC_24, _MCW_PC );
	double divSpeedReg = DivTest("dividing nrm by nrm", 22, 1.12341234, 0.0, kMinCount * 10 );
	DivTest("dividing nrm by 16", 22, 16, divSpeedReg, kMinCount * 10 );
	DivTest("dividing nrm by zero", 22, zero, divSpeedReg, kMinCount );

	AddHeading( "Divide tests - double precision" );
	_controlfp( _PC_53, _MCW_PC );
	DivTest("dividing nrm by nrm", 22, 1.12341234, divSpeedReg, kMinCount * 10 );
	DivTest("dividing nrm by 16", 22, 16, divSpeedReg, kMinCount * 10 );
	DivTest("dividing nrm by zero", 22, zero, divSpeedReg, kMinCount );

	AddHeading( "Divide tests - extended precision" );
	_controlfp( _PC_64, _MCW_PC );
	DivTest("dividing nrm by nrm", 22, 1.12341234, divSpeedReg, kMinCount * 10 );
	DivTest("dividing nrm by 16", 22, 16, divSpeedReg, kMinCount * 10 );
	DivTest("dividing nrm by zero", 22, zero, divSpeedReg, kMinCount );

	AddHeading( "Multiply tests - float precision" );
	_controlfp( _PC_24, _MCW_PC );
	double mulSpeedReg = MulTest("multiply nrm by nrm", 22, 1.12341234, 0.0, kMinCount * 10 );
	MulTest("multiply nrm by 16", 22, 16, mulSpeedReg, kMinCount * 10 );
	MulTest("multiply to overflow", DBL_MAX, 2, mulSpeedReg, kMinCount );
	MulTest("multiply to underflow", nrm, .25, mulSpeedReg, kMinCount );

	AddHeading( "Multiply tests - double precision" );
	_controlfp( _PC_53, _MCW_PC );
	MulTest("multiply nrm by nrm", 22, 1.12341234, mulSpeedReg, kMinCount * 10 );
	MulTest("multiply nrm by 16", 22, 16, mulSpeedReg, kMinCount * 10 );
	MulTest("multiply to overflow", DBL_MAX, 2, mulSpeedReg, kMinCount );
	MulTest("multiply to underflow", nrm, .25, mulSpeedReg, kMinCount );

	AddHeading( "Multiply tests - extended precision" );
	_controlfp( _PC_64, _MCW_PC );
	MulTest("multiply nrm by nrm", 22, 1.12341234, mulSpeedReg, kMinCount * 10 );
	MulTest("multiply nrm by 16", 22, 16, mulSpeedReg, kMinCount * 10 );
	MulTest("multiply to overflow", DBL_MAX, 2, mulSpeedReg, kMinCount );
	MulTest("multiply to underflow", nrm, .25, mulSpeedReg, kMinCount );

	if (sse2)
	{
		AddHeading( "Multiply tests - SSE2" );
		MulTestSSE2("multiply nrm by nrm", 22, 1.12341234, mulSpeedReg, kMinCount * 10 );
		MulTestSSE2("multiply nrm by 16", 22, 16, mulSpeedReg, kMinCount * 10 );
		MulTestSSE2("multiply to overflow", DBL_MAX, 2, mulSpeedReg, kMinCount );
		MulTestSSE2("multiply to underflow", nrm, .25, mulSpeedReg, kMinCount );
	}

	AddHeading( "Square root tests - float precision" );
	_controlfp( _PC_24, _MCW_PC );
	double sqrtSpeedReg = SqrtTest("sqrt nrm", 1.12343213, 0.0, kMinCount );
	SqrtTest("sqrt 4", 4, sqrtSpeedReg, kMinCount * 5 );
	SqrtTest("sqrt 2", 2, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt 9", 9, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt negative", -1.12343213, sqrtSpeedReg, kMinCount );

	AddHeading( "Square root tests - double precision" );
	_controlfp( _PC_53, _MCW_PC );
	SqrtTest("sqrt nrm", 1.12343213, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt 4", 4, sqrtSpeedReg, kMinCount * 5 );
	SqrtTest("sqrt 2", 2, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt 9", 9, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt negative", -1.12343213, sqrtSpeedReg, kMinCount );

	AddHeading( "Square root tests - extended precision" );
	_controlfp( _PC_64, _MCW_PC );
	SqrtTest("sqrt nrm", 1.12343213, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt 4", 4, sqrtSpeedReg, kMinCount * 5 );
	SqrtTest("sqrt 2", 2, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt 9", 9, sqrtSpeedReg, kMinCount );
	SqrtTest("sqrt negative", -1.12343213, sqrtSpeedReg, kMinCount );

	// Reset precision to double for remaining tests
	_controlfp( _PC_53, _MCW_PC );
#endif	// DO_FULL_MATH_TESTS


	AddHeading( "Register to register adds" );
	double addSpeedReg = AddTestReg("adding nrm to nrm", nrm, nrm, 0.0, kMinCount * 100 );
	AddTestReg( "adding nan to nan", nan, nan, addSpeedReg );
	AddTestReg( "adding inf to inf", inf, inf, addSpeedReg );
	AddTestReg( "adding inf to nan", nan, inf, addSpeedReg );
	AddTestReg( "adding nrm to inf", inf, nrm, addSpeedReg );
	AddTestReg( "adding inf to nrm", zero, inf, addSpeedReg );
	AddTestReg( "adding den to nrm", nrm, den, addSpeedReg, kMinCount * 100 );

	if (sse2)
	{
		AddHeading( "SSE2 adds" );
		double addSSE2SpeedReg = AddTestSSE2("adding nrm to nrm", nrm, nrm, 0.0, kMinCount * 100 );
		AddTestSSE2( "adding nan to nan", nan, nan, addSSE2SpeedReg, kMinCount * 10 );
		AddTestSSE2( "adding inf to inf", inf, inf, addSSE2SpeedReg, kMinCount * 10 );
		AddTestSSE2( "adding inf to nan", nan, inf, addSSE2SpeedReg, kMinCount * 10 );
		AddTestSSE2( "adding nrm to inf", inf, nrm, addSSE2SpeedReg, kMinCount * 10 );
		AddTestSSE2( "adding inf to nrm", zero, inf, addSSE2SpeedReg, kMinCount * 10 );
		AddTestSSE2( "adding den to nrm", nrm, den, addSSE2SpeedReg, kMinCount * 10 );
	}

	AddHeading( "Memory to register adds" );
	double addSpeedMem = AddTestMem( "adding nrm to nrm", nrm, nrm, 0, kMinCount * 50 );
	AddTestMem( "adding den to nrm", nrm, den, addSpeedMem );

	AddHeading( "Loads and stores - type tests" );
	double loadStoreSpeedNormal = LoadTest( "Loading nrm", nrm, nrm, 0, 0, 0.0, kMinCount * 50 );
	LoadTest( "Loading nan", nrm, nan, 0, 0, loadStoreSpeedNormal, kMinCount * 5 );
	LoadTest( "Loading inf", nrm, inf, 0, 0, loadStoreSpeedNormal, kMinCount * 50 );
	LoadTest( "Loading den", nrm, den, 0, 0, loadStoreSpeedNormal, kMinCount );

	AddHeading( "Loads and stores - alignment tests" );
	double loadStoreSpeedAligned = LoadTest( "quad aligned", nrm, nrm, kAlignSize - 8, 0, 0.0, kMinCount * 50 );
	LoadTest( "dword aligned src", nrm, nrm, kAlignSize - 12, 0, loadStoreSpeedAligned, kMinCount * 50 );
	LoadTest( "byte aligned src", nrm, nrm, kAlignSize - 11, 0, loadStoreSpeedAligned, kMinCount * 50 );
	LoadTest( "cache unaligned src", nrm, nrm, 124, 0, loadStoreSpeedAligned, kMinCount * 2 );
	LoadTest( "page unaligned src", nrm, nrm, kAlignSize - 4, 0, loadStoreSpeedAligned, kMinCount * 2 );
	LoadTest( "dword aligned dst", nrm, nrm, 0, kAlignSize - 12, loadStoreSpeedAligned, kMinCount * 50 );
	LoadTest( "byte aligned dst", nrm, nrm, 0, kAlignSize - 11, loadStoreSpeedAligned, kMinCount * 50 );
	LoadTest( "cache unaligned dst", nrm, nrm, 0, 124, loadStoreSpeedAligned, kMinCount * 2 );
	LoadTest( "page unaligned dst", nrm, nrm, 0, kAlignSize - 4, loadStoreSpeedAligned, kMinCount * 2 );
}
