#include "stdafx.h"

#include "support.h"
#include <windows.h>
#include <cmath>
#include <cassert>

struct TimeEntry
{
	__int64	rdtscTime;
	double	worldTime;
};




__int64 GetCycleCount()
{
	__int64 result;
	__asm
	{
		push	eax
		push	edx
		rdtsc
		mov		DWORD PTR result, eax
		mov		DWORD PTR result + 4, edx
		pop		edx
		pop		eax
	}
	return result;
}




static double GetSeconds()
{
	static bool Initialized;
	LARGE_INTEGER	iFrequency, iTime;
	BOOL FreqResult = QueryPerformanceFrequency(&iFrequency);
	BOOL CounterResult = QueryPerformanceCounter(&iTime);
	if ((FreqResult == TRUE) && (CounterResult == TRUE))
	{
		double Time = (double)iTime.QuadPart / (double)iFrequency.QuadPart;
		return Time;
	}
	assert( 0 );
	return 0;
}





static TimeEntry GetTimeEntry()
{
	int OldPriority = GetThreadPriority(GetCurrentThread());

	__int64 sStartTime;
	// Record start clock count and time. Use high priority to reduce
	// the chance of losing our time slice partway through.
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
	__asm
	{
		push	eax
		push	edx
		rdtsc
		mov		DWORD PTR sStartTime, eax
		mov		DWORD PTR sStartTime + 4, edx
		pop		edx
		pop		eax
	}
	TimeEntry	result;
	result.worldTime = GetSeconds();
	SetThreadPriority(GetCurrentThread(), OldPriority);
	result.rdtscTime = sStartTime;

	return result;
}




// Returns the approximate CPU speed in Hz. It is returned in a double
// because an unsigned int will be too small very soon.
// This method typically gives results that are consistent, and presumably
// accurate, to five decimal places.
double GetCPUSpeed()
{
	static bool initialized = false;
	static double s_cpuSpeed;
	if ( !initialized )
	{
		initialized = true;
		TimeEntry start = GetTimeEntry();
		TimeEntry end;
		// We now want to pause for a short time, and then compare the
		// elapsed time to the change in rdtsc. We could do a Sleep(200),
		// which would work on most computers, but on some laptops the
		// CPU would go to sleep, and the rdtsc counter would stop
		// incrementing, which would make it a poor measure of CPU speed.
		double pauseLength = 0.2;
		while ( true )
		{
			end = GetTimeEntry();
			if ( end.worldTime > start.worldTime + pauseLength )
				break;
		}
		TimeEntry delta;
		delta.rdtscTime = end.rdtscTime - start.rdtscTime;
		delta.worldTime = end.worldTime - start.worldTime;
		// Calculate the CPU speed
		s_cpuSpeed = delta.rdtscTime / delta.worldTime;
		// Round to the nearest MHz, to give more consistent results.
		s_cpuSpeed = floor( ( s_cpuSpeed / 1000000 ) + .5 ) * 1000000;
	}

	return s_cpuSpeed;
}




static void PrintMainHeading()
{
	static bool s_firstTime = true;
	if ( s_firstTime )
	{
		s_firstTime = false;
		printf( "Testing on %1.1f MHz CPU\n", GetCPUSpeed() / 1000000 );
		printf( "%-21s    count      time    Mops/sec  Cycles/op  Slowdown\n", "Identifier" );
	}
}




// This function prints out the Timings from the test. It can be easily
// modified to save the results to a tab or comma separated file also.
void AddTimings( const char* identifier, int count, double elapsedTime, double baseSpeed )
{
	PrintMainHeading();

	double slowdown = 1;
	if ( baseSpeed )
	{
		double thisSpeed = count / elapsedTime;
		slowdown = baseSpeed / thisSpeed;
	}
	printf( "%-21s, %9d, %8.5f, %8.3f, %8.3f, %8.3f\n", identifier, count, elapsedTime,
				(count / elapsedTime) / 1000000,
				GetCPUSpeed() / (count / elapsedTime),
				slowdown );
}




void AddHeading( const char* heading )
{
	PrintMainHeading();

	printf( "\n\t%s\n", heading );
}




// Do nothing function to stop the optimizer from removing calculations completely.
void SetResult( double result )
{
}


bool HasSSE2()
{
	bool available = true;

	// Set up an exception handler so we don't crash if
	// the check fails
	int	CPUIDResults = 0;
	__try
	{
		// Detecting SSE2 the Intel way, as described in the
		// IA-32 Intel(R) Architecture Software Developer's Manual, volume 1
        // The description is in section 11.6.2: Checking for SSE and SSE2 support
		__asm
		{
			push	eax
			push	ebx
			push	ecx
			push	edx
			// First, execute CPUID. If this instruction doesn't exist, we crash,
			// it's caught by the exception handler, and we continue.
			mov		eax, 1
			cpuid
			// Bit 25 says that SSE is supported and bit 26 says that SSE2 is
			// supported. Intel also says to check bit 24 for FXSAVE and a few
			// other bits, but I don't care about those.
            // Curiously, the Intel manual doesn't clearly say what register these
            // bit values are put in - but with a bit of cross referencing you can
            // determine that it is edx.
			mov		CPUIDResults, edx

			pop		edx
			pop		ecx
			pop		ebx
			pop		eax
		}
        int mask = 1 << 26;
		if ( !(CPUIDResults & mask) )
			available = false;
	}
	__except( EXCEPTION_EXECUTE_HANDLER )
	{
		// If we got here, then the SSE instructions caused an exception
		// and we can assume the system does not support SSE.
		available = false;
	}

	return available;
}
